# 11 "src/session/strategy_parser.mll"
 
  open Strategy

  exception SyntaxError of string

  let error f =
    Printf.kbprintf
      (fun b ->
        let s = Buffer.contents b in Buffer.clear b; raise (SyntaxError s))
      (Buffer.create 1024)
      f

  type label = {
    mutable defined: int option;
    temporary: int;
  }

  type 'a code = {
    env: Env.env;
    whyconf: Whyconf.config;
    mutable instr: instruction array;
    mutable next: int;
    labels: (string, label) Hashtbl.t; (* label name -> label *)
    mutable temp: int;
  }

  let create_code env conf =
    let h = Hashtbl.create 17 in
    Hashtbl.add h "exit" { defined = Some (-1); temporary = 0 };
    { env = env;
      whyconf = conf;
      instr = Array.make 10 (Igoto 0);
      next = 0;
      temp = 1;
      labels = h; }

  let enlarge_code code =
    let old = code.instr in
    let n = Array.length old in
    code.instr <- Array.make (2 * n) (Igoto 0);
    Array.blit old 0 code.instr 0 n

  let add_instr code i =
    let n = code.next in
    if n = Array.length code.instr then enlarge_code code;
    code.instr.(n) <- i;
    code.next <- n + 1

  let temp code =
    let t = code.temp in code.temp <- t + 1; t

  let define_label code l =
    let n = code.next in
    try
      let lab = Hashtbl.find code.labels l in
      if lab.defined = None then lab.defined <- Some n
      else error "duplicate label %s" l
    with Not_found ->
      let lab =  { defined = Some n; temporary = temp code } in
      Hashtbl.add code.labels l lab

  let find_label code l =
    try
      let lab = Hashtbl.find code.labels l in lab.temporary
    with Not_found ->
      let t = temp code in
      Hashtbl.add code.labels l { defined = None; temporary = t };
      t

  let prover code p =
    try
      let fp = Whyconf.parse_filter_prover p in
      Whyconf.filter_one_prover code.whyconf fp
    with
      | Whyconf.ProverNotFound _ ->
          error "Prover %S not installed or not configured" p
      | Whyconf.ProverAmbiguity _ ->
          error "Prover description %s is ambiguous" p

  let integer msg s =
    try Some (int_of_string s)
    with Failure _ ->
      match s with
      | "%t" -> None
      | "%.t" -> None
      | "%T" -> None
      | "%m" -> None
      | "%s" -> None
      | _ ->
          error "unable to parse %s argument '%s'" msg s

  let real msg s =
    try Some (float_of_string s)
    with Failure _ ->
      match s with
      | "%t" -> None
      | "%.t" -> None
      | "%T" -> None
      | "%S" -> None
      | "%m" -> None
      | "%s" -> None
      | _ ->
          error "unable to parse %s argument '%s'" msg s


  let transform code t =
    try
      ignore (Trans.lookup_transform t code.env)
    with Trans.UnknownTrans _ ->
    try
      ignore (Trans.lookup_transform_l t code.env)
    with Trans.UnknownTrans _->
      error "transformation %S is unknown" t


# 118 "src/session/strategy_parser.ml"
let __ocaml_lex_tables = {
  Lexing.lex_base =
   "\000\000\002\000\016\000\000\000\042\000\069\000\081\000\131\000\
    \000\000\141\000\190\000\007\000\013\000\216\000\018\000\009\000\
    \076\000\139\000\146\000\192\000\161\000\242\000\227\000\221\000\
    \023\000\020\001\248\255\249\255\035\001\037\001\044\001\050\001\
    \004\000\052\001\254\255\032\000\064\001\253\255\077\001\079\001\
    \094\001\107\001\115\001\121\001\135\001\145\001\150\001\159\001\
    \165\001\174\001\179\001\203\001\212\001\217\001\232\001\245\001\
    \250\001\023\002\030\002";
  Lexing.lex_backtrk =
   "\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\000\000\255\255\255\255\255\255\000\000\255\255\255\255\
    \000\000\000\000\000\000\000\000\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\006\000\006\000\006\000\006\000\
    \006\000\000\000\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\003\000\004\000\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\005\000";
  Lexing.lex_default =
   "\001\000\001\000\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\028\000\000\000\000\000\038\000\038\000\038\000\038\000\
    \035\000\255\255\000\000\035\000\042\000\000\000\038\000\038\000\
    \038\000\038\000\042\000\255\255\038\000\038\000\038\000\056\000\
    \038\000\038\000\038\000\038\000\038\000\038\000\038\000\038\000\
    \056\000\058\000\058\000";
  Lexing.lex_trans =
   "\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\255\255\255\255\002\000\002\000\255\255\034\000\002\000\
    \018\000\018\000\016\000\016\000\018\000\000\000\016\000\000\000\
    \000\000\002\000\002\000\014\000\014\000\002\000\000\000\014\000\
    \255\255\000\000\002\000\255\255\000\000\255\255\000\000\018\000\
    \000\000\016\000\034\000\000\000\000\000\000\000\024\000\000\000\
    \002\000\000\000\014\000\007\000\007\000\003\000\000\000\007\000\
    \000\000\000\000\255\255\006\000\255\255\006\000\004\000\000\000\
    \005\000\005\000\005\000\005\000\005\000\005\000\005\000\005\000\
    \005\000\005\000\007\000\000\000\000\000\000\000\007\000\007\000\
    \000\000\000\000\007\000\000\000\023\000\016\000\016\000\000\000\
    \000\000\016\000\004\000\004\000\004\000\004\000\004\000\004\000\
    \004\000\004\000\004\000\004\000\000\000\007\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\016\000\019\000\000\000\020\000\
    \000\000\000\000\000\000\004\000\023\000\005\000\005\000\005\000\
    \005\000\005\000\005\000\005\000\005\000\005\000\005\000\004\000\
    \017\000\006\000\006\000\006\000\006\000\006\000\006\000\006\000\
    \006\000\006\000\006\000\023\000\007\000\007\000\015\000\020\000\
    \007\000\000\000\000\000\000\000\014\000\014\000\010\000\010\000\
    \014\000\000\000\010\000\018\000\018\000\000\000\000\000\018\000\
    \000\000\000\000\000\000\007\000\000\000\000\000\000\000\000\000\
    \008\000\000\000\000\000\014\000\000\000\010\000\000\000\000\000\
    \000\000\000\000\018\000\009\000\009\000\009\000\009\000\009\000\
    \009\000\009\000\009\000\009\000\009\000\009\000\009\000\009\000\
    \009\000\009\000\009\000\009\000\009\000\009\000\009\000\010\000\
    \010\000\010\000\010\000\010\000\022\000\010\000\022\000\000\000\
    \000\000\021\000\021\000\021\000\021\000\021\000\021\000\021\000\
    \021\000\021\000\021\000\000\000\000\000\000\000\010\000\000\000\
    \010\000\014\000\014\000\012\000\000\000\014\000\007\000\007\000\
    \000\000\000\000\007\000\000\000\000\000\000\000\013\000\013\000\
    \013\000\013\000\013\000\013\000\013\000\013\000\013\000\013\000\
    \014\000\000\000\000\000\007\000\007\000\007\000\000\000\007\000\
    \255\255\000\000\255\255\000\000\034\000\000\000\000\000\000\000\
    \013\000\013\000\013\000\013\000\013\000\013\000\013\000\013\000\
    \013\000\013\000\007\000\021\000\021\000\021\000\021\000\021\000\
    \021\000\021\000\021\000\021\000\021\000\033\000\033\000\000\000\
    \034\000\033\000\021\000\021\000\021\000\021\000\021\000\021\000\
    \021\000\021\000\021\000\021\000\255\255\255\255\047\000\047\000\
    \255\255\000\000\047\000\000\000\033\000\043\000\043\000\032\000\
    \000\000\043\000\011\000\036\000\036\000\033\000\033\000\036\000\
    \000\000\033\000\000\000\255\255\000\000\047\000\255\255\000\000\
    \255\255\036\000\036\000\000\000\043\000\036\000\027\000\255\255\
    \000\000\000\000\036\000\000\000\033\000\255\255\255\255\255\255\
    \255\255\255\255\255\255\000\000\255\255\037\000\000\000\037\000\
    \036\000\000\000\000\000\255\255\000\000\000\000\037\000\255\255\
    \255\255\000\000\000\000\255\255\037\000\255\255\000\000\255\255\
    \255\255\000\000\255\255\000\000\036\000\036\000\000\000\030\000\
    \036\000\000\000\255\255\031\000\255\255\255\255\255\255\000\000\
    \255\255\255\255\043\000\043\000\000\000\000\000\043\000\037\000\
    \029\000\037\000\000\000\036\000\000\000\044\000\255\255\000\000\
    \255\255\255\255\000\000\255\255\255\255\000\000\255\255\048\000\
    \037\000\043\000\255\255\255\255\000\000\000\000\255\255\043\000\
    \043\000\039\000\000\000\043\000\000\000\037\000\000\000\255\255\
    \047\000\047\000\255\255\000\000\047\000\255\255\255\255\255\255\
    \000\000\255\255\255\255\000\000\255\255\000\000\043\000\255\255\
    \255\255\255\255\000\000\255\255\255\255\255\255\000\000\047\000\
    \255\255\037\000\255\255\040\000\000\000\255\255\000\000\000\000\
    \255\255\000\000\000\000\037\000\000\000\041\000\255\255\000\000\
    \037\000\255\255\000\000\255\255\255\255\255\255\255\255\000\000\
    \255\255\255\255\000\000\000\000\000\000\255\255\255\255\037\000\
    \000\000\255\255\255\255\255\255\000\000\000\000\255\255\000\000\
    \037\000\000\000\000\000\255\255\000\000\037\000\255\255\000\000\
    \000\000\255\255\255\255\045\000\255\255\255\255\000\000\255\255\
    \000\000\255\255\000\000\000\000\255\255\046\000\047\000\047\000\
    \000\000\000\000\047\000\057\000\057\000\037\000\049\000\057\000\
    \255\255\000\000\000\000\255\255\000\000\000\000\037\000\000\000\
    \000\000\000\000\000\000\037\000\026\000\047\000\000\000\000\000\
    \255\255\000\000\057\000\000\000\050\000\255\255\000\000\000\000\
    \057\000\057\000\037\000\255\255\057\000\255\255\051\000\255\255\
    \255\255\000\000\000\000\255\255\255\255\000\000\000\000\037\000\
    \000\000\052\000\255\255\000\000\255\255\000\000\000\000\057\000\
    \000\000\000\000\255\255\000\000\000\000\000\000\255\255\000\000\
    \255\255\255\255\000\000\053\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\054\000\000\000\255\255\000\000\255\255\
    \000\000\255\255\000\000\000\000\000\000\055\000\000\000\000\000\
    \255\255\000\000\000\000\000\000\000\000\000\000\255\255\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\255\255\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\255\255\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\255\255\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\255\255\000\000\000\000\000\000\000\000\255\255\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\255\255\
    \000\000\000\000\000\000\000\000\000\000\255\255\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\255\255\000\000\
    \000\000\000\000\000\000\255\255\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\255\255\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\255\255\000\000\000\000\000\000\
    \000\000\255\255\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \255\255\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\255\255\000\000\000\000\
    \000\000\000\000\255\255\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\255\255\
    \000\000\000\000\000\000\000\000\000\000\000\000\255\255";
  Lexing.lex_check =
   "\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\000\000\000\000\001\000\001\000\000\000\032\000\001\000\
    \011\000\011\000\015\000\015\000\011\000\255\255\015\000\255\255\
    \255\255\002\000\002\000\014\000\014\000\002\000\255\255\014\000\
    \000\000\255\255\001\000\000\000\255\255\001\000\255\255\011\000\
    \255\255\015\000\035\000\255\255\255\255\255\255\003\000\255\255\
    \002\000\255\255\014\000\004\000\004\000\002\000\255\255\004\000\
    \255\255\255\255\000\000\002\000\001\000\002\000\002\000\255\255\
    \002\000\002\000\002\000\002\000\002\000\002\000\002\000\002\000\
    \002\000\002\000\004\000\255\255\255\255\255\255\005\000\005\000\
    \255\255\255\255\005\000\255\255\003\000\016\000\016\000\255\255\
    \255\255\016\000\004\000\004\000\004\000\004\000\004\000\004\000\
    \004\000\004\000\004\000\004\000\255\255\005\000\255\255\255\255\
    \255\255\255\255\255\255\255\255\016\000\008\000\255\255\004\000\
    \255\255\255\255\255\255\005\000\003\000\005\000\005\000\005\000\
    \005\000\005\000\005\000\005\000\005\000\005\000\005\000\006\000\
    \012\000\006\000\006\000\006\000\006\000\006\000\006\000\006\000\
    \006\000\006\000\006\000\024\000\007\000\007\000\014\000\004\000\
    \007\000\255\255\255\255\255\255\017\000\017\000\009\000\009\000\
    \017\000\255\255\009\000\018\000\018\000\255\255\255\255\018\000\
    \255\255\255\255\255\255\007\000\255\255\255\255\255\255\255\255\
    \007\000\255\255\255\255\017\000\255\255\009\000\255\255\255\255\
    \255\255\255\255\018\000\007\000\007\000\007\000\007\000\007\000\
    \007\000\007\000\007\000\007\000\007\000\009\000\009\000\009\000\
    \009\000\009\000\009\000\009\000\009\000\009\000\009\000\010\000\
    \010\000\019\000\019\000\010\000\020\000\019\000\020\000\255\255\
    \255\255\020\000\020\000\020\000\020\000\020\000\020\000\020\000\
    \020\000\020\000\020\000\255\255\255\255\255\255\010\000\255\255\
    \019\000\013\000\013\000\010\000\255\255\013\000\023\000\023\000\
    \255\255\255\255\023\000\255\255\255\255\255\255\010\000\010\000\
    \010\000\010\000\010\000\010\000\010\000\010\000\010\000\010\000\
    \013\000\255\255\255\255\021\000\021\000\023\000\255\255\021\000\
    \000\000\255\255\001\000\255\255\032\000\255\255\255\255\255\255\
    \013\000\013\000\013\000\013\000\013\000\013\000\013\000\013\000\
    \013\000\013\000\021\000\022\000\022\000\022\000\022\000\022\000\
    \022\000\022\000\022\000\022\000\022\000\025\000\025\000\255\255\
    \035\000\025\000\021\000\021\000\021\000\021\000\021\000\021\000\
    \021\000\021\000\021\000\021\000\028\000\028\000\029\000\029\000\
    \028\000\255\255\029\000\255\255\025\000\030\000\030\000\025\000\
    \255\255\030\000\010\000\031\000\031\000\033\000\033\000\031\000\
    \255\255\033\000\255\255\028\000\255\255\029\000\028\000\255\255\
    \029\000\036\000\036\000\255\255\030\000\036\000\025\000\030\000\
    \255\255\255\255\031\000\255\255\033\000\031\000\038\000\038\000\
    \039\000\039\000\038\000\255\255\039\000\028\000\255\255\029\000\
    \036\000\255\255\255\255\036\000\255\255\255\255\030\000\040\000\
    \040\000\255\255\255\255\040\000\031\000\038\000\255\255\039\000\
    \038\000\255\255\039\000\255\255\041\000\041\000\255\255\025\000\
    \041\000\255\255\036\000\025\000\042\000\042\000\040\000\255\255\
    \042\000\040\000\043\000\043\000\255\255\255\255\043\000\038\000\
    \025\000\039\000\255\255\041\000\255\255\030\000\041\000\255\255\
    \044\000\044\000\255\255\042\000\044\000\255\255\042\000\029\000\
    \040\000\043\000\045\000\045\000\255\255\255\255\045\000\046\000\
    \046\000\031\000\255\255\046\000\255\255\041\000\255\255\044\000\
    \047\000\047\000\044\000\255\255\047\000\042\000\048\000\048\000\
    \255\255\045\000\048\000\255\255\045\000\255\255\046\000\049\000\
    \049\000\046\000\255\255\049\000\050\000\050\000\255\255\047\000\
    \050\000\044\000\047\000\039\000\255\255\048\000\255\255\255\255\
    \048\000\255\255\255\255\045\000\255\255\040\000\049\000\255\255\
    \046\000\049\000\255\255\050\000\051\000\051\000\050\000\255\255\
    \051\000\047\000\255\255\255\255\255\255\052\000\052\000\048\000\
    \255\255\052\000\053\000\053\000\255\255\255\255\053\000\255\255\
    \049\000\255\255\255\255\051\000\255\255\050\000\051\000\255\255\
    \255\255\054\000\054\000\044\000\052\000\054\000\255\255\052\000\
    \255\255\053\000\255\255\255\255\053\000\045\000\055\000\055\000\
    \255\255\255\255\055\000\056\000\056\000\051\000\048\000\056\000\
    \054\000\255\255\255\255\054\000\255\255\255\255\052\000\255\255\
    \255\255\255\255\255\255\053\000\025\000\055\000\255\255\255\255\
    \055\000\255\255\056\000\255\255\049\000\056\000\255\255\255\255\
    \057\000\057\000\054\000\028\000\057\000\029\000\050\000\058\000\
    \058\000\255\255\255\255\058\000\030\000\255\255\255\255\055\000\
    \255\255\051\000\031\000\255\255\056\000\255\255\255\255\057\000\
    \255\255\255\255\057\000\255\255\255\255\255\255\058\000\255\255\
    \036\000\058\000\255\255\052\000\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\053\000\255\255\038\000\255\255\039\000\
    \255\255\057\000\255\255\255\255\255\255\054\000\255\255\255\255\
    \058\000\255\255\255\255\255\255\255\255\255\255\040\000\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\041\000\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\042\000\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\044\000\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\045\000\255\255\255\255\255\255\255\255\046\000\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\047\000\
    \255\255\255\255\255\255\255\255\255\255\048\000\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\049\000\255\255\
    \255\255\255\255\255\255\050\000\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\051\000\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\052\000\255\255\255\255\255\255\
    \255\255\053\000\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \054\000\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\055\000\255\255\255\255\
    \255\255\255\255\056\000\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\057\000\
    \255\255\255\255\255\255\255\255\255\255\255\255\058\000";
  Lexing.lex_base_code =
   "\000\000\002\000\013\000\000\000\039\000\069\000\000\000\095\000\
    \000\000\121\000\147\000\000\000\013\000\173\000\007\000\000\000\
    \000\000\015\000\000\000\041\000\158\000\222\000\188\000\067\000\
    \013\000\000\000\000\000\000\000\000\000\101\000\000\000\126\000\
    \000\000\000\000\000\000\000\000\152\000\000\000\000\000\000\000\
    \000\000\237\000\000\000\000\000\000\000\000\000\000\000\250\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\252\000\
    \021\001\023\001\000\000";
  Lexing.lex_backtrk_code =
   "\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\016\000\000\000\000\000\000\000\039\000\000\000\000\000\
    \059\000\039\000\076\000\016\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\097\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\106\000";
  Lexing.lex_default_code =
   "\001\000\001\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\100\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \100\000\000\000\000\000";
  Lexing.lex_trans_code =
   "\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\004\000\004\000\000\000\000\000\004\000\
    \056\000\056\000\000\000\000\000\056\000\000\000\004\000\004\000\
    \056\000\056\000\004\000\000\000\056\000\000\000\000\000\000\000\
    \000\000\000\000\004\000\000\000\000\000\000\000\000\000\056\000\
    \000\000\000\000\000\000\000\000\000\000\004\000\007\000\056\000\
    \010\000\010\000\031\000\031\000\010\000\000\000\031\000\000\000\
    \000\000\000\000\000\000\007\000\000\000\007\000\007\000\007\000\
    \007\000\007\000\007\000\007\000\007\000\007\000\007\000\010\000\
    \000\000\031\000\000\000\000\000\010\000\010\000\010\000\010\000\
    \010\000\000\000\010\000\000\000\007\000\000\000\000\000\007\000\
    \007\000\007\000\007\000\007\000\007\000\007\000\007\000\007\000\
    \007\000\000\000\000\000\010\000\000\000\010\000\000\000\000\000\
    \010\000\010\000\000\000\000\000\010\000\013\000\091\000\091\000\
    \000\000\000\000\091\000\007\000\007\000\007\000\007\000\007\000\
    \007\000\007\000\007\000\007\000\007\000\007\000\007\000\010\000\
    \036\000\007\000\031\000\031\000\000\000\091\000\031\000\094\000\
    \094\000\000\000\000\000\094\000\000\000\000\000\000\000\013\000\
    \013\000\013\000\013\000\013\000\013\000\013\000\013\000\013\000\
    \013\000\031\000\000\000\000\000\031\000\031\000\094\000\000\000\
    \031\000\094\000\094\000\000\000\000\000\094\000\000\000\000\000\
    \000\000\013\000\013\000\013\000\013\000\013\000\013\000\013\000\
    \013\000\013\000\013\000\031\000\000\000\000\000\056\000\056\000\
    \094\000\000\000\056\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\036\000\036\000\036\000\036\000\036\000\
    \036\000\036\000\036\000\036\000\036\000\056\000\007\000\007\000\
    \007\000\007\000\007\000\007\000\007\000\007\000\007\000\007\000\
    \000\000\000\000\000\000\000\000\000\000\036\000\036\000\036\000\
    \036\000\036\000\036\000\036\000\036\000\036\000\036\000\010\000\
    \010\000\000\000\000\000\010\000\007\000\007\000\007\000\007\000\
    \007\000\007\000\007\000\007\000\007\000\007\000\094\000\094\000\
    \000\000\000\000\094\000\000\000\000\000\000\000\010\000\000\000\
    \000\000\000\000\000\000\091\000\091\000\091\000\091\000\091\000\
    \000\000\091\000\000\000\000\000\000\000\094\000\007\000\007\000\
    \007\000\007\000\007\000\007\000\007\000\007\000\007\000\007\000\
    \000\000\000\000\091\000\000\000\091\000\000\000\103\000\103\000\
    \103\000\103\000\103\000\000\000\103\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\103\000\000\000\103\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    ";
  Lexing.lex_check_code =
   "\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\000\000\000\000\001\000\001\000\000\000\255\255\001\000\
    \014\000\014\000\255\255\255\255\014\000\255\255\002\000\002\000\
    \017\000\017\000\002\000\255\255\017\000\255\255\255\255\255\255\
    \000\000\255\255\001\000\000\000\255\255\001\000\255\255\014\000\
    \255\255\255\255\255\255\255\255\255\255\002\000\006\000\017\000\
    \004\000\004\000\019\000\019\000\004\000\255\255\019\000\255\255\
    \255\255\255\255\000\000\002\000\001\000\002\000\002\000\002\000\
    \002\000\002\000\002\000\002\000\002\000\002\000\002\000\004\000\
    \255\255\019\000\255\255\255\255\023\000\023\000\005\000\005\000\
    \023\000\255\255\005\000\255\255\003\000\255\255\255\255\004\000\
    \004\000\004\000\004\000\004\000\004\000\004\000\004\000\004\000\
    \004\000\255\255\255\255\023\000\255\255\005\000\255\255\255\255\
    \007\000\007\000\255\255\255\255\007\000\008\000\029\000\029\000\
    \255\255\255\255\029\000\005\000\003\000\005\000\005\000\005\000\
    \005\000\005\000\005\000\005\000\005\000\005\000\005\000\007\000\
    \012\000\024\000\009\000\009\000\255\255\029\000\009\000\031\000\
    \031\000\255\255\255\255\031\000\255\255\255\255\255\255\007\000\
    \007\000\007\000\007\000\007\000\007\000\007\000\007\000\007\000\
    \007\000\009\000\255\255\255\255\010\000\010\000\031\000\255\255\
    \010\000\036\000\036\000\255\255\255\255\036\000\255\255\255\255\
    \255\255\009\000\009\000\009\000\009\000\009\000\009\000\009\000\
    \009\000\009\000\009\000\010\000\255\255\255\255\013\000\013\000\
    \036\000\255\255\013\000\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\010\000\010\000\010\000\010\000\010\000\
    \010\000\010\000\010\000\010\000\010\000\013\000\020\000\020\000\
    \020\000\020\000\020\000\020\000\020\000\020\000\020\000\020\000\
    \255\255\255\255\255\255\255\255\255\255\013\000\013\000\013\000\
    \013\000\013\000\013\000\013\000\013\000\013\000\013\000\021\000\
    \021\000\255\255\255\255\021\000\022\000\022\000\022\000\022\000\
    \022\000\022\000\022\000\022\000\022\000\022\000\041\000\041\000\
    \255\255\255\255\041\000\255\255\255\255\255\255\021\000\255\255\
    \000\000\255\255\001\000\047\000\047\000\055\000\055\000\047\000\
    \255\255\055\000\255\255\255\255\255\255\041\000\021\000\021\000\
    \021\000\021\000\021\000\021\000\021\000\021\000\021\000\021\000\
    \255\255\255\255\047\000\255\255\055\000\047\000\056\000\056\000\
    \057\000\057\000\056\000\255\255\057\000\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\047\000\056\000\255\255\057\000\
    \056\000\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\056\000\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\047\000\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\056\000\255\255\255\255\
    ";
  Lexing.lex_code =
   "\255\008\255\255\009\255\255\010\255\255\011\255\255\012\255\255\
    \007\255\006\255\001\009\002\010\000\008\003\011\004\012\255\014\
    \255\013\255\255\015\255\255\007\255\001\009\002\010\006\013\005\
    \015\000\008\003\011\004\012\255\014\255\255\001\009\002\010\007\
    \014\006\013\005\015\000\008\003\011\004\012\255\006\255\001\009\
    \002\010\007\014\000\008\003\011\004\012\255\004\255\255\003\255\
    \255\000\003\255\005\255\255\006\255\255\000\004\001\005\002\006\
    \255";
}

let rec invok code lexbuf =
  lexbuf.Lexing.lex_mem <- Array.make 16 (-1); __ocaml_lex_invok_rec code lexbuf 0
and __ocaml_lex_invok_rec code lexbuf __ocaml_lex_state =
  match Lexing.new_engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
let
# 142 "src/session/strategy_parser.mll"
            p
# 533 "src/session/strategy_parser.ml"
= Lexing.sub_lexeme lexbuf lexbuf.Lexing.lex_start_pos lexbuf.Lexing.lex_mem.(0)
and
# 142 "src/session/strategy_parser.mll"
                                    t
# 538 "src/session/strategy_parser.ml"
= Lexing.sub_lexeme lexbuf lexbuf.Lexing.lex_mem.(1) lexbuf.Lexing.lex_mem.(2)
and
# 142 "src/session/strategy_parser.mll"
                                                           m
# 543 "src/session/strategy_parser.ml"
= Lexing.sub_lexeme lexbuf lexbuf.Lexing.lex_mem.(3) lexbuf.Lexing.lex_mem.(4)
and
# 143 "src/session/strategy_parser.mll"
                               s
# 548 "src/session/strategy_parser.ml"
= Lexing.sub_lexeme_opt lexbuf lexbuf.Lexing.lex_mem.(6) lexbuf.Lexing.lex_mem.(5)
and
# 143 "src/session/strategy_parser.mll"
                                                    sep
# 553 "src/session/strategy_parser.ml"
= Lexing.sub_lexeme_char_opt lexbuf lexbuf.Lexing.lex_mem.(7) in
# 144 "src/session/strategy_parser.mll"
  (
    let p = prover code p in
    let t = real "timelimit" t in
    if t <> None && Option.get t <= 0.0 then
      error "timelimit %f is invalid" (Option.get t);
    let m = integer "memlimit" m in
    if m <> None && Option.get m <= 0 then
      error "memlimit %d is invalid" (Option.get m);
    let s = integer "steplimit" (Option.value ~default:"%s" s) in
    if s <> None && Option.get s < 0 then
      error "steplimit %d is invalid" (Option.get s);
    if sep <> None then
      (p.Whyconf.prover, t, m, s) :: invok code lexbuf
    else
      [p.Whyconf.prover, t, m, s]
  )
# 572 "src/session/strategy_parser.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf;
      __ocaml_lex_invok_rec code lexbuf __ocaml_lex_state

and scan code lexbuf =
  lexbuf.Lexing.lex_mem <- Array.make 7 (-1); __ocaml_lex_scan_rec code lexbuf 25
and __ocaml_lex_scan_rec code lexbuf __ocaml_lex_state =
  match Lexing.new_engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
# 162 "src/session/strategy_parser.mll"
      ( scan code lexbuf )
# 584 "src/session/strategy_parser.ml"

  | 1 ->
# 164 "src/session/strategy_parser.mll"
      ( scan code lexbuf )
# 589 "src/session/strategy_parser.ml"

  | 2 ->
let
# 165 "src/session/strategy_parser.mll"
             id
# 595 "src/session/strategy_parser.ml"
= Lexing.sub_lexeme lexbuf lexbuf.Lexing.lex_start_pos (lexbuf.Lexing.lex_curr_pos + -1) in
# 166 "src/session/strategy_parser.mll"
      ( define_label code id;
        scan code lexbuf )
# 600 "src/session/strategy_parser.ml"

  | 3 ->
let
# 168 "src/session/strategy_parser.mll"
                          id
# 606 "src/session/strategy_parser.ml"
= Lexing.sub_lexeme lexbuf lexbuf.Lexing.lex_mem.(0) lexbuf.Lexing.lex_curr_pos in
# 169 "src/session/strategy_parser.mll"
      ( add_instr code (Igoto (find_label code id));
        scan code lexbuf )
# 611 "src/session/strategy_parser.ml"

  | 4 ->
# 172 "src/session/strategy_parser.mll"
    ( let i = invok code lexbuf in
      add_instr code (Icall_prover i);
      scan code lexbuf )
# 618 "src/session/strategy_parser.ml"

  | 5 ->
let
# 175 "src/session/strategy_parser.mll"
                               t
# 624 "src/session/strategy_parser.ml"
= Lexing.sub_lexeme lexbuf lexbuf.Lexing.lex_mem.(0) lexbuf.Lexing.lex_mem.(1)
and
# 175 "src/session/strategy_parser.mll"
                                                   l
# 629 "src/session/strategy_parser.ml"
= Lexing.sub_lexeme lexbuf lexbuf.Lexing.lex_mem.(2) lexbuf.Lexing.lex_curr_pos in
# 176 "src/session/strategy_parser.mll"
      ( transform code t;
        add_instr code (Itransform (t, find_label code l));
        scan code lexbuf )
# 635 "src/session/strategy_parser.ml"

  | 6 ->
let
# 179 "src/session/strategy_parser.mll"
         c
# 641 "src/session/strategy_parser.ml"
= Lexing.sub_lexeme_char lexbuf lexbuf.Lexing.lex_start_pos in
# 180 "src/session/strategy_parser.mll"
      ( let i = Lexing.lexeme_start lexbuf in
        error "syntax error on character '%c' at position %d" c i )
# 646 "src/session/strategy_parser.ml"

  | 7 ->
# 183 "src/session/strategy_parser.mll"
      ( () )
# 651 "src/session/strategy_parser.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf;
      __ocaml_lex_scan_rec code lexbuf __ocaml_lex_state

;;

# 185 "src/session/strategy_parser.mll"
 

  let parse env conf s =
    let code = create_code env conf in
    scan code (Lexing.from_string s);
    let label = Array.make code.temp 0 in
    let fill name lab = match lab.defined with
      | None -> error "label '%s' is undefined" name
      | Some n -> label.(lab.temporary) <- n in
    Hashtbl.iter fill code.labels;
    let solve = function
      | Icall_prover _ as i -> i
      | Itransform (t, n) -> Itransform (t, label.(n))
      | Igoto n -> Igoto label.(n) in
    Array.map solve (Array.sub code.instr 0 code.next)


# 676 "src/session/strategy_parser.ml"
